#######
# assemble.py
# description : assemble archived & unarchived urls to reconstruct a Web site
# licence : AGPL + CECILL v3
# author : quentin lobbé - qlobbe@iscpif.fr
#######

import sys
import csv
import datetime
import collections
import json
from bs4 import BeautifulSoup
import urllib.request

#    archived :: path to csv
#  unarchived :: path to csv
archived    = ""
unarchived  = ""

#   time_unit :: year | month
# time_period :: int
#   time_step :: int
time_unit   = "month"
time_period = 3
time_step   = 1

try :
    archived   = sys.argv[1]
    unarchived = sys.argv[2]
except :
    print ("! args error\n")
    print ("try : python3 archivesToFragments.py output/archives-firsttuesday.com-1996-2010.csv output/unarchived-1996-2010.csv\n")
    sys.exit(0)


# archive {
#   url :: string,
#   timestamp :: [int],
#   time_unit :: [int],
#   periods :: [(date,date)],
#   status  :: bool
#   angle   :: float
#   depth   :: int
# }

# create monthly time_unit from a timestamp
def toMonth (ts,is_date) :

    tmp  = datetime.datetime(1,1,1)
    if is_date :
        date = ts
    else :
        date = datetime.datetime(int(ts[0:4]),int(ts[4:6]),int(ts[6:8]))
    
    return (date.year - tmp.year) * 12 + (date.month - tmp.month)


# read a csv of archived/unarchived links and create a corpus
def readCsv (path,status) :

    with open(path, newline='\n') as file :

        raw_data = csv.reader(file, delimiter=';')
        next(raw_data, None)
    
        for row in raw_data :

            try :
                int(row[0])
            except :
                continue
            
            if row[2] not in corpus :
                corpus[row[2]] = {
                    "url" : row[2],
                    "timestamp" : [int(row[0])],
                    "time_unit" : [toMonth(row[0],False)],
                    "periods" : [],
                    "status" : status,
                    "angle" : 0,
                    "depth" : 0
                }
            else :
                corpus[row[2]]["timestamp"].append(int(row[0]))
                corpus[row[2]]["time_unit"].append(toMonth(row[0],False))


def getTimeUnitFrags (frags) :
    arr = []
    for frag in frags :
       arr.append(frag["time_unit"])
    return arr    

def getTimeUnitCorpus (cps) :
    arr = []
    for k, v in cps.items() :
       arr += v["time_unit"]
    return arr    

def getDepth (cps) :
    arr = []
    for k, v in cps.items() :
        arr.append(v["depth"])
    return arr


# init the periods
def toPeriods (min_date,max_date,tu,tp,ts) :

    if tu != "month" :
        print ("'month' is the only time_unit implemented so far\n")
        sys.exit(0)

    prds = []
    curr = min_date

    while curr <= max_date :
        prds.append((curr, curr + tp - 1))
        curr += ts

    return prds


# is a time_unit contained in a period ?
def isInPeriod(tu,prd) :
    if tu >= prd[0] and tu <= prd[1] :
        return True
    else :
        return False


# attach periods to the corpus 
def indexPeriods () :

    for k in corpus :
        for tu in corpus[k]["time_unit"] :
            for prd in periods :
                if isInPeriod(tu,prd) :
                    corpus[k]["periods"].append(prd) 


# attach angles to the sorted corpus
def indexAngles () :
    
    i = 0
    inc = 360 / (len(corpus) + 1)
    for k, v in corpus.items() :
        v["angle"] = i * inc
        i += 1


# attach depths to the sorted corpus
def indexDepths () :
    
    for k, v in corpus.items() :
        tmp = ((v["url"].replace("http://","")).split("?"))[0]
        v["depth"] = tmp.count('/') + 1


# clean periods and time_unit params
def cleanCorpus () :
    
    for k, v in corpus.items() :
        v["periods"] = list(set(v["periods"]))
        v["time_unit"] = list(set(v["time_unit"])) 
        v["timestamp"] = list(set(v["timestamp"]))       


def formatDate(ts,date,format) :

    if date is None or date == "" :
        tu = toMonth(str(ts),False)
    elif format == 1 :
        try :
            tu = toMonth(datetime.datetime.strptime(date, '%H:%M  %A, %B %d, %Y'),True)
        except :
            tu = toMonth(str(ts),False)
    else :
        tu = None

    return tu;


def getArchive(url) :
    try :
        req = urllib.request.Request(url)
        with urllib.request.urlopen(req) as res:
            return res.read()
    except Exception as e :
        print(str(e))
        return None  


def cleanText (content) :
    return content.replace('\r',' ').replace('\t',' ').replace('\n',' ').replace('"','')


def getContent(soup,pattern) :
    # scrape the content of a fragment
    content = "NA"
    try :
        if   pattern == 1 :
            content = soup.div.get_text()
        elif pattern == 2 :
            content = soup.find_all('table')[2].get_text()
        elif pattern in {3,4} :
            content = soup.find_all('p')[0].get_text()
        elif pattern in {5} :
            if "#0033CC" not in str(soup[0]) and "<br/>" in str(soup[0]) :
                content = soup[0].get_text()
            elif "#0033CC" not in str(soup[1]) and "<br/>" in str(soup[1]) :
                content = soup[1].get_text() 
            elif "#0033CC" not in str(soup[2]) and "<br/>" in str(soup[2]) :
                content = soup[2].get_text()
        elif pattern in {6} :
            els = soup.find_all('font')
            content = ""
            for el in els :
                content += el.get_text() + " "
        elif pattern in {7} :
            content = soup.find_all("td", class_="main")[0].text.strip() 
        elif pattern in {8} :
            content = soup.find_all("td")[7].text.strip()                      
        return cleanText(content)      
    except Exception as e :
        return content


def getTitle (soup,pattern) :
    # scrape the title of a fragment
    title = "NA"
    try :
        if   pattern in {1,2,9} :
            title = soup.find_all('eventtitle')[0].string
        elif pattern in {3,4} :
            title = soup.find_all('b')[1].get_text()
        elif pattern in {5} :
            if "#0033CC" in str(soup[0]) and "Location" not in str(soup[0]):
                title = soup[0].text.strip()
            elif "#0033CC" in str(soup[1]) and "Location" not in str(soup[1]):
                title = soup[1].text.strip()     
            elif "#0033CC" in str(soup[2]) and "Location" not in str(soup[2]):
                title = soup[2].text.strip()
        elif pattern in {6} :
            title = soup.find_all('b')[0].text.strip()
        elif pattern in {7} :
            title = soup.find_all("div", class_="smtitle")[0].text.strip()  
        elif pattern in {8} :
            title = soup.find_all('b')[3].text.strip()               
        return cleanText(title)
    except Exception as e :
        return title


def getDate (soup,ts,pattern) :
    # scrape the date of a fragment
    date = "NA"
    try :
        if   pattern in {1,2} :
            date = (soup.find_all('eventdate'))[0].string
        elif pattern in {3} :
            date = (soup.find_all('b')[2].next_sibling).replace(": ","").replace("\n","")
        elif pattern in {4} :
            date = (((soup.find_all('b'))[0].string).split('in')[0]).strip()
            date += ' ' + str(ts)[0:4]
        elif pattern in {5} :
            if len(soup[0].find_all('b')) == 1 and "  in " in (soup[0].find_all('b')[0]).text :
                date = (((soup[0].find_all('b')[0]).text.split("  in "))[0]).strip()
            elif len(soup[1].find_all('b')) == 1 and "  in " in (soup[1].find_all('b')[0]).text :
                date = (((soup[1].find_all('b')[0]).text.split("  in "))[0]).strip()
            elif len(soup[2].find_all('b')) == 1 and "  in " in (soup[2].find_all('b')[0]).text :
                date = (((soup[2].find_all('b')[0]).text.split("  in "))[0]).strip()
            elif "Date" in soup[0].text :
                date = (soup[0].find_all('b')[0].nextSibling).replace(":","").strip()
            elif "Date" in soup[1].text :
                date = (soup[1].find_all('b')[0].nextSibling).replace(":","").strip()          

        return date
    except Exception as e :
        return date


def getPlace (soup,url,pattern) :
    # scrape the place of a fragment
    place = "NA"
    try :
        if   pattern in {1,2} :
            place = (soup.find_all('eventdateb'))[0].string
        elif pattern in {3,4} :
            place = (((soup.find_all('b'))[0].string).split('in')[1]).strip()
        elif pattern in {5} :
            if len(soup[0].find_all('b')) == 1 and "  in " in (soup[0].find_all('b')[0]).text :
                place = (((soup[0].find_all('b')[0]).text.split("  in "))[1]).strip()
            elif len(soup[1].find_all('b')) == 1 and "  in " in (soup[1].find_all('b')[0]).text :
                place = (((soup[1].find_all('b')[0]).text.split("  in "))[1]).strip() 
            elif len(soup[2].find_all('b')) == 1 and "  in " in (soup[2].find_all('b')[0]).text :
                place = (((soup[2].find_all('b')[0]).text.split("  in "))[1]).strip()                                
            elif "Location" in soup[0].text :
                place = (soup[0].find_all('b')[1].nextSibling).replace(":","").strip()
            elif "Location" in soup[1].text :
                place = (soup[1].find_all('b')[1].nextSibling).replace(":","").strip()
        elif pattern in {8} :  
                place = url.split('/')[4]
                if place == "" :
                    place = "NA"            
        return place
    except Exception as e :
        return place 


def getEventId (url) :
    # extract the event id of a fragment
    if "id=" in url :
        return url.split('id=')[1]
    else :
        return "NA"            


def toTimeUnit (date,ts,pattern) :
    # turn a scraped date into a time unit
    try : 
        if date == "NA" :
            return toMonth(str(ts),False)
        elif pattern in {1,2} :
            return toMonth(datetime.datetime.strptime(date, '%H:%M  %A, %B %d, %Y'),True)
        elif pattern in {3} :
            return toMonth(datetime.datetime.strptime(date.split(' at ')[0], '%d-%m-%Y'),True)
        elif pattern in {4} :
            return toMonth(datetime.datetime.strptime(date, '%d %B %Y'),True)  
        elif pattern in {5} :
            if " at " not in date :
                return toMonth(datetime.datetime.strptime(str(ts)[0:4] + ' ' + date, '%Y %d %B'), True)
            else :
                return toMonth(datetime.datetime.strptime(date.split(" at ")[0], '%d-%m-%Y'), True)

    except :
        return toMonth(str(ts),False)  

def toFragment(ts,url,pattern,wayback,soup) :  

    title = getTitle(soup,pattern)
    place = getPlace(soup,url,pattern)
    date  = getDate(soup,ts,pattern)
    event_id  = getEventId(url)
    content   = getContent(soup,pattern)
    time_unit = toTimeUnit(date,ts,pattern)

    source_angle = corpus[url]["angle"]
    source_depth = corpus[url]["depth"]
    source_periods = corpus[url]["periods"]

    # print(title)

    fragments.append({
            'url' : url,
            'title' : title,
            'date' : date,
            'place' : place,
            'time_unit' : time_unit,
            'content' : content,
            'wayback' : wayback,
            'source_angle'   : source_angle,
            'source_depth'   : source_depth,
            'source_periods' : source_periods
        })       


def getFragment(ts,url,pattern) :
    
    id = str(ts) + 'id_'
    
    wayback = "https://web.archive.org/web/" + id + "/" + url.replace(':80/','/')
    page = getArchive(wayback)
    
    print("########################\n")
    print(wayback)
    print("########################\n")
    
    if page is None or page == "" :
   
        print("Unable to fragment " + url)
   
    else :
        
        soup = BeautifulSoup(page, 'html.parser')

        if pattern in {5} :

            lst = []
            for el in soup.find_all('font') :
                if ("ff3300" not in str(el)) or ("Verdana" in str(el)):
                    lst.append(el)
            triplets = [lst[i:i + 3] for i in range(0, len(lst), 3)]
            
            for triplet in triplets :
                if len(triplet) != 3 :
                    continue
                toFragment(ts,url,pattern,wayback,triplet)

        else :
            toFragment(ts,url,pattern,wayback,soup)


# fragment {
#   url :: string,
#   title :: string,
#   date :: string,
#   place :: string,
#   time_unit :: int
#   content :: string,
# }



def toFragments() :

    i = 0

    for k, v in corpus.items() :

        mask = k.replace("www.","").replace(":80","").lower()
        # print(mask)

        if v["status"] :
        
            if mask.startswith("http://firsttuesday.com/newerster/code/eventsignup.asp?description=1&eid=") :
                # continue
                getFragment(v["timestamp"][0],k,1)
            
            elif mask.startswith("http://firsttuesday.com/code/eventsignup.asp?description=1&eid=") :
                # continue
                getFragment(v["timestamp"][0],k,2)

            elif mask.startswith("http://firsttuesday.com/eventdetail.asp?id=") :
                # continue
                getFragment(v["timestamp"][0],k,3)

            elif mask.startswith("http://firsttuesday.com/cgi-local/full_news.cgi?event_id=") :
                # continue
                getFragment(v["timestamp"][0],k,4)

            elif mask.startswith("http://firsttuesday.com/eventdetail.asp") :

                for ts in v["timestamp"] :
                    # continue
                    getFragment(ts,k,5)

            elif mask.startswith("http://firsttuesday.com/main/events_news") or mask.startswith("http://firsttuesday.com/content/events_news/") :
                getFragment(v["timestamp"][0],k,6)
                # continue

            elif mask.startswith("http://firsttuesday.com/html/content/features/money_talks") :
                getFragment(v["timestamp"][0],k,7)
                # continue

            elif mask.startswith("http://firsttuesday.com/wirelesswednesday/") :
                getFragment(v["timestamp"][0],k,8)
                # continue



            
             

##
# main code
##

# corpus :: {url:archive}
corpus    = {}
periods   = [] 
fragments = []

readCsv(archived,True)
readCsv(unarchived,False)

# periods :: [(date,date)]
periods = toPeriods(min(getTimeUnitCorpus(corpus))
                   ,max(getTimeUnitCorpus(corpus)),time_unit,time_period,time_step)
indexPeriods()

# sort the corpus of archives
corpus = collections.OrderedDict(sorted(corpus.items()))
indexAngles()
indexDepths()
cleanCorpus()

toFragments()

with open('assembled_archive.json', 'w') as file :
   json.dump({'periods' : periods, 
              'corpus'  : corpus, 
              'max_depth' : max(getDepth(corpus)), 
              'min_date'  : min([min(getTimeUnitFrags(fragments)),min(getTimeUnitCorpus(corpus))]), 
              'max_date'  : max([max(getTimeUnitFrags(fragments)),max(getTimeUnitCorpus(corpus))]), 
              'fragments' : fragments}, file, sort_keys=True, indent=4)